import dataclasses


@dataclasses.dataclass
class CalibrationBoard:
    rows: int
    columns: int
    square_size: float
    x_offset: float
    y_offset: float


CalibrationBoard8by6 = CalibrationBoard(
    rows=6,
    columns=8,
    square_size=0.031,
    x_offset=0.16,
    y_offset=-0.124,
)



from typing import List

import cv2

from .boards import CalibrationBoard
from ... import BGRImage, Pixel


def find_corners(image: BGRImage, board: CalibrationBoard) -> List[Pixel]:
    """
    Finds the corners of the given board in the given image.

    Note:
        The provided image should be rectified.

    Args:
        image (:obj:``numpy array``): A color (3-channel) OpenCV image
        board (:obj:``dt_computer_vision.camera.calibration.extrinsics.boards.CalibrationBoard``):
            The calibration board to look for

    Returns:
        :obj:``list[dt_computer_vision.camera.Pixel]``: Corners detected in the given image,
        ordered left-right, top-bottom.

    Raises:
        RuntimeError: If no corners were found in image, or the corners couldn't be rearranged

    """
    grayscale = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)
    corners_num = (board.columns - 1, board.rows - 1)

    # find corners in the image
    ret, corners = cv2.findChessboardCorners(grayscale, corners_num, cv2.CALIB_CB_ADAPTIVE_THRESH)
    if not ret:
        raise RuntimeError("No corners found in image, or the corners couldn't be rearranged. "
                           "Make sure that the camera is positioned correctly.")

    # refine corners' position
    criteria = (cv2.TERM_CRITERIA_EPS + cv2.TERM_CRITERIA_MAX_ITER, 100, 0.001)
    corners = cv2.cornerSubPix(grayscale,
                               corners,
                               winSize=(11, 11),
                               zeroZone=(-1, -1),
                               criteria=criteria)

    # return corners as Pixels
    corners = [Pixel(c[0], c[1]) for c in corners[:, 0]]
    # ---
    return corners



"""

    line_detector
    -------------

    The ``line_detector`` library packages classes and tools for handling line section extraction from images. The
    main functionality is in the :py:class:`LineDetector` class. :py:class:`Detections` is the output data class for
    the results of a call to :py:class:`LineDetector`, and :py:class:`ColorRange` is used to specify the colour ranges
    in which :py:class:`LineDetector` is looking for line segments.

    There are two plotting utilities also included: :py:func:`plotMaps` and :py:func:`plotSegments`

    .. autoclass:: line_detector.Detections

    .. autoclass:: line_detector.ColorRange

    .. autoclass:: line_detector.LineDetector

    .. autofunction:: line_detector.plotMaps

    .. autofunction:: line_detector.plotSegments


"""

from .types import Detections, ColorRange
from .line_detector import LineDetector



import dataclasses
from typing import Optional, Tuple

import cv2
import numpy as np

from .utils import invert_map, ensure_ndarray

BGRImage = np.ndarray
HSVImage = np.ndarray


@dataclasses.dataclass
class Point:
    """
    A generic 2D point.
    """
    x: float
    y: float

    def as_array(self) -> np.ndarray:
        return np.array([self.x, self.y])

    def __repr__(self):
        return f"P({self.x}, {self.y})"


class Pixel(Point):

    def as_integers(self) -> np.ndarray:
        return np.array([int(self.x), int(self.y)], dtype=int)


class NormalizedImagePoint(Point):
    pass


class Rectifier:
    """
    Handles the Rectification operations.

    """

    def __init__(self, camera: 'CameraModel'):
        self.camera: CameraModel = camera
        self._rectify_inited: bool = False
        self._distort_inited: bool = False
        # forward maps (lazily initialized)
        self.mapx: Optional[np.ndarray] = None
        self.mapy: Optional[np.ndarray] = None
        # inverse maps (lazily initialized)
        self.rmapx: Optional[np.ndarray] = None
        self.rmapy: Optional[np.ndarray] = None

    def _init_rectify_maps(self):
        W = self.camera.width
        H = self.camera.height
        mapx = np.ndarray(shape=(H, W, 1), dtype="float32")
        mapy = np.ndarray(shape=(H, W, 1), dtype="float32")
        self.mapx, self.mapy = cv2.initUndistortRectifyMap(
            self.camera.K, self.camera.D, self.camera.R, self.camera.P,
            (W, H), cv2.CV_32FC1, mapx, mapy
        )
        self._rectify_inited = True

    def rectify_pixel(self, point: Pixel) -> Pixel:
        """
        Args:
            point (:py:class:`Pixel`): A point in pixel coordinates.

        Applies the rectification specified by camera parameters
        :math:`K` and and :math:`D` to the given pixel (u, v) and returns the
        pixel coordinates of the rectified point.
        """
        src = point.as_array().reshape((1, 1, 2)).astype(float)
        dst = cv2.undistortPoints(src,
                                  self.camera.K,
                                  self.camera.D,
                                  R=self.camera.R,
                                  P=self.camera.P)
        return Pixel(*dst[0, 0])

    def rectify(self, image: BGRImage, interpolation=cv2.INTER_NEAREST):
        """
        Undistorts an image.
        While cv2.INTER_NEAREST is faster, use cv2.INTER_CUBIC for higher quality.
        """
        if not self._rectify_inited:
            self._init_rectify_maps()
        # rectify distorted image
        return cv2.remap(image, self.mapx, self.mapy, interpolation)

    def distort(self, rectified: BGRImage, interpolation=cv2.INTER_NEAREST) -> np.ndarray:
        """
        Distorts an image.
        While cv2.INTER_NEAREST is faster, use cv2.INTER_CUBIC for higher quality.
        """
        # initialize forward map (if not done yet)
        if not self._rectify_inited:
            self._init_rectify_maps()
        # initialize inverse map (if not done yet)
        if not self._distort_inited:
            self.rmapx, self.rmapy = invert_map(self.mapx, self.mapy)
            self._distort_inited = True
        # distort rectified image
        return cv2.remap(rectified, self.rmapx, self.rmapy, interpolation)


@dataclasses.dataclass
class CameraModel:
    width: int
    height: int
    K: np.ndarray
    D: np.ndarray
    P: np.ndarray
    R: np.ndarray = None
    H: Optional[np.ndarray] = None

    rectifier: Rectifier = dataclasses.field(init=False)

    def __post_init__(self):
        self.K = ensure_ndarray(self.K)
        self.D = ensure_ndarray(self.D)
        self.P = ensure_ndarray(self.P)
        self.R = np.eye(3) if self.R is None else ensure_ndarray(self.R)
        self.H = None if self.H is None else ensure_ndarray(self.H)
        self.rectifier = Rectifier(self)

    @property
    def fx(self) -> float:
        return self.K[0, 0]

    @property
    def fy(self) -> float:
        return self.K[1, 1]

    @property
    def cx(self) -> float:
        return self.K[0, 2]

    @property
    def cy(self) -> float:
        return self.K[1, 2]

    def get_shape(self) -> Tuple[int, int]:
        """returns (height, width) of image"""
        return self.height, self.width

    def vector2pixel(self, vec: NormalizedImagePoint) -> Pixel:
        """
        Converts a ``[-1,1] X [-1,1]`` representation to ``[0, W] X [0, H]``
        (from normalized to image coordinates).

        Args:
            vec (:py:class:`Point`): A :py:class:`Point` object in normalized coordinates.

        Returns:
            :py:class:`Point` : A :py:class:`Point` object in image coordinates.

        """
        x = self.cx + vec.x * self.fx
        y = self.cy + vec.y * self.fy
        return Pixel(x, y)

    def pixel2vector(self, pixel: Pixel) -> NormalizedImagePoint:
        """
        Converts a ``[0,W] X [0,H]`` representation to ``[-1, 1] X [-1, 1]``
        (from image to normalized coordinates).

        Args:
            pixel (:py:class:`Point`): A :py:class:`Point` object in image coordinates.

        Returns:
            :py:class:`Point` : A :py:class:`Point` object in normalized coordinates.

        """
        x = (pixel.x - self.cx) / self.fx
        y = (pixel.y - self.cy) / self.fy
        return NormalizedImagePoint(x, y)



from typing import List

import numpy as np
import cv2

from .boards import CalibrationBoard
from ... import BGRImage, Pixel


def draw_corners(image: BGRImage, board: CalibrationBoard, corners: List[Pixel]) -> BGRImage:
    """
    Draws the given corners on the given image.

    Note:
        The provided image should be rectified.

    Args:
        image (:obj:``numpy array``): A color (3-channel) OpenCV image
        board (:obj:``dt_computer_vision.camera.calibration.extrinsics.boards.CalibrationBoard``):
            The calibration board to look for
        corners: (:obj:``list[Pixel]``): Corners as returned by
        ``dt_computer_vision.camera.calibration.extrinsics.chessboard.find_corners``

    Returns:
        :obj:``numpy array``: The estimated homography

    Raises:
        RuntimeError: If no corners were found in image, or the corners couldn't be rearranged

    """
    grayscale = cv2.cvtColor(cv2.cvtColor(image, cv2.COLOR_BGR2GRAY), cv2.COLOR_GRAY2BGR)
    corners_num = (board.columns - 1, board.rows - 1)
    # opencv wants corners in an array of shape [N, 1, 2]
    corners = np.expand_dims([c.as_array() for c in corners], axis=1)
    # draw chessboard on image
    return cv2.drawChessboardCorners(grayscale, corners_num, corners, True)



import math
from typing import List, Tuple, Union, Optional

import cv2
import numpy as np

from dt_computer_vision.camera import BGRImage

Color = Union[Tuple[float], List[float]]


class AntiInstagram:

    def __init__(self,
                 image_scale: float = 0.2,
                 color_balance_scale: float = 0.8,
                 lower_threshold: Optional[Color] = None,
                 higher_threshold: Optional[Color] = None):
        # store parameters
        self.color_balance_scale = color_balance_scale
        self.image_scale = image_scale
        self.lower_threshold = lower_threshold or [0, 0, 0]
        self.higher_threshold = higher_threshold or [255, 255, 255]

    def update(self, image: BGRImage):
        # scale image (if necessary)
        if self.image_scale != 1.0:
            image = cv2.resize(image, (0, 0), fx=self.image_scale, fy=self.image_scale)

        # TODO: this cropping seems arbitrary and Duckiebot-specific, if needed, pass it as a `roi` argument
        # H = image.shape[0]
        # image = image[int(H * 0.3): (H - 1), :, :]

        half_percent = self.color_balance_scale / 2
        channels = cv2.split(image)

        lower_threshold = []
        for idx, channel in enumerate(channels):
            # find the low and high percentile values (based on the input percentile)
            height, width = channel.shape
            num_pixels = width * height
            flattened = channel.reshape(num_pixels)
            # sort entries
            flattened = np.sort(flattened)
            # calculate thresholds
            lower_threshold.append(flattened[int(math.floor(num_pixels * half_percent))])
        # update lower threshold
        self.lower_threshold = lower_threshold

    def apply(self,
              image: BGRImage,
              image_scale: float = 1.0,
              lower_threshold: Optional[Color] = None,
              higher_threshold: Optional[Color] = None):
        lower_thr = lower_threshold or self.lower_threshold
        higher_thr = higher_threshold or self.higher_threshold
        # scale image (if necessary)
        if image_scale != 1.0:
            image = cv2.resize(image, (0, 0), fx=image_scale, fy=image_scale)
        # split BGR -> B, G, R
        channels = cv2.split(image)
        out_channels = []
        # saturate each channel
        for idx, channel in enumerate(channels):
            thresholded = self._apply_threshold(channel, lower_thr[idx], higher_thr[idx])
            # TODO: that `.copy()` might be expensive
            normalized = cv2.normalize(thresholded, thresholded.copy(), 0, 255, cv2.NORM_MINMAX)
            out_channels.append(normalized)
        # recombine B, G, R -> BGR
        return cv2.merge(out_channels)

    @classmethod
    def _apply_threshold(cls, channel, low_value, high_value):
        # create mask for pixels below threshold
        low_mask = channel < low_value
        # saturate highlighted pixels with low value
        channel = cls._apply_mask(channel, low_mask, low_value)
        # create mask for pixels above threshold
        high_mask = channel > high_value
        # saturate highlighted pixels with high value
        channel = cls._apply_mask(channel, high_mask, high_value)
        # return saturated channel
        return channel

    @classmethod
    def _apply_mask(cls, matrix, mask, fill_value):
        masked = np.ma.array(matrix, mask=mask, fill_value=fill_value)
        return masked.filled()



from abc import abstractmethod, ABC
from typing import Tuple, List

import cv2
import numpy as np

from .types import Detections, ColorRange
from ..camera import BGRImage, HSVImage


class LineDetectorInterface(ABC):

    @abstractmethod
    def detect(self, image: BGRImage, color: ColorRange) -> Detections:
        """Returns a tuple of class Detections"""
        pass


class LineDetector(LineDetectorInterface):
    """
    The Line Detector can be used to extract line segments from a particular color range in
    an image. It combines edge detection, color filtering, and line segment extraction.

    This class was created for the goal of extracting the white, yellow, and red lines in the
    Duckiebot's camera stream as part of the lane localization pipeline. It is setup in a way
    that allows efficient detection of line segments in different color ranges.

    In order to process an image, first the :py:meth:`setImage` method must be called.
    In makes an internal copy of the image, converts it to `HSV color space
    <https://en.wikipedia.org/wiki/HSL_and_HSV>`_, which is much better for color segmentation,
    and applies `Canny edge detection <https://en.wikipedia.org/wiki/Canny_edge_detector>`_.

    Then, to do the actual line segment extraction, a call to :py:meth:`detect` with a
    :py:class:`ColorRange` object must be made.
    Multiple such calls with different colour ranges can be made and these will reuse the
    precomputed HSV image and Canny edges.

    Args:

        canny_thresholds (:obj:`list` of :obj:`int`): a list with two entries that specify the
        thresholds for the hysteresis procedure, details `here
        <https://docs.opencv.org/2.4/modules/imgproc/doc/feature_detection.html?highlight=canny
        #canny>`__, default is ``[80, 200]``

        canny_aperture_size (:obj:`int`): aperture size for a Sobel operator, details `here
        <https://docs.opencv.org/2.4/modules/imgproc/doc/feature_detection.html?highlight=canny
        #canny>`__, default is 3

        dilation_kernel_size (:obj:`int`): kernel size for the dilation operation which fills
        in the gaps in the color filter result, default is 3

        hough_threshold (:obj:`int`): Accumulator threshold parameter. Only those lines are
        returned that get enough votes, details `here
        <https://docs.opencv.org/2.4/modules/imgproc/doc/feature_detection.html?highlight
        =houghlinesp#houghlinesp>`__, default is 2

        hough_min_line_length (:obj:`int`): Minimum line length. Line segments shorter than that
        are rejected, details `here
        <https://docs.opencv.org/2.4/modules/imgproc/doc/feature_detection.html?highlight
        =houghlinesp#houghlinesp>`__, default is 3

        hough_max_line_gap (:obj:`int`): Maximum allowed gap between points on the same line to
        link them, details `here
        <https://docs.opencv.org/2.4/modules/imgproc/doc/feature_detection.html?highlight
        =houghlinesp#houghlinesp>`__, default is 1

    """

    def __init__(
            self,
            canny_thresholds=None,
            canny_aperture_size=3,
            dilation_kernel_size=3,
            hough_threshold=2,
            hough_min_line_length=3,
            hough_max_line_gap=1,
    ):
        self.canny_thresholds = canny_thresholds or [80, 200]
        self.canny_aperture_size = canny_aperture_size
        self.dilation_kernel_size = dilation_kernel_size
        self.hough_threshold = hough_threshold
        self.hough_min_line_length = hough_min_line_length
        self.hough_max_line_gap = hough_max_line_gap

    @staticmethod
    def find_edges(image: BGRImage, canny_thr_low: int, canny_thr_high: int,
                   canny_aperture_size: int) -> np.ndarray:
        """
        Applies `Canny edge detection <https://en.wikipedia.org/wiki/Canny_edge_detector>`_
        to a ``BGR`` image.

        Args:
            image: (:obj:`numpy array`): BGR image to process.

            canny_thr_low (:obj:`int`): low threshold for the hysteresis procedure, details `here
            <https://docs.opencv.org/2.4/modules/imgproc/doc/feature_detection.html?highlight=canny
            #canny>`__.

            canny_thr_high (:obj:`int`): high threshold for the hysteresis procedure, details `here
            <https://docs.opencv.org/2.4/modules/imgproc/doc/feature_detection.html?highlight=canny
            #canny>`__.

            canny_aperture_size (:obj:`int`): aperture size for a Sobel operator, details `here
            <https://docs.opencv.org/2.4/modules/imgproc/doc/feature_detection.html?highlight=canny
            #canny>`__.

        Returns:
            :obj:`numpy array`: a binary image with the edges
        """
        # perform edge detection
        edges = cv2.Canny(
            image,
            canny_thr_low,
            canny_thr_high,
            apertureSize=canny_aperture_size,
        )
        return edges

    @staticmethod
    def hough_line(edges: np.ndarray, hough_threshold: int,
                   hough_min_line_length: int,
                   hough_max_line_gap: int, ) -> np.ndarray:
        """
        Finds line segments in a binary image using the probabilistic Hough transform.
        Based on the OpenCV function
        `HoughLinesP
        <https://docs.opencv.org/2.4/modules/imgproc/doc/feature_detection.html?highlight=houghlinesp#houghlinesp>`_.

        Args:
            edges (:obj:`numpy array`): binary image with edges

            hough_threshold (:obj:`int`): Accumulator threshold parameter. Only those lines are
            returned that get enough votes, details `here
            <https://docs.opencv.org/2.4/modules/imgproc/doc/feature_detection.html?highlight
            =houghlinesp#houghlinesp>`__.

            hough_min_line_length (:obj:`int`): Minimum line length. Line segments shorter than
            that are rejected, details `here
            <https://docs.opencv.org/2.4/modules/imgproc/doc/feature_detection.html?highlight
            =houghlinesp#houghlinesp>`__.

            hough_max_line_gap (:obj:`int`): Maximum allowed gap between points on the same line to
            link them, details `here
            <https://docs.opencv.org/2.4/modules/imgproc/doc/feature_detection.html?highlight
            =houghlinesp#houghlinesp>`__.

        Returns:
             :obj:`numpy array`: An ``Nx4`` array where each row represents a
             line ``[x1, y1, x2, y2]``. If no lines
             were detected, returns an empty list.

        """
        lines = cv2.HoughLinesP(
            edges,
            rho=1,
            theta=np.pi / 180,
            threshold=hough_threshold,
            minLineLength=hough_min_line_length,
            maxLineGap=hough_max_line_gap,
        )
        if lines is not None:
            lines = lines.reshape((-1, 4))  # it has an extra dimension
        else:
            lines = np.array([])

        return lines

    @staticmethod
    def color_filter(image: HSVImage, color: ColorRange, edges: np.ndarray,
                     dilation_kernel_size: int) -> Tuple[np.ndarray, np.ndarray]:
        """
        Obtains the regions of the image that fall in the provided color range and the subset
        of the detected Canny edges which are in these regions.
        Applies a `dilation <https://homepages.inf.ed.ac.uk/rbf/HIPR2/dilate.htm>`_
        operation to smooth and grow the regions map.

        Args:
            image: (:obj:`numpy array`): HSV image to process.

            color (:py:class:`ColorRange`): A :py:class:`ColorRange` object specifying
            the desired colors.

            edges (:obj:`numpy array`): binary image with the edges as returned by ``find_edges``.

            dilation_kernel_size (:obj:`int`): kernel size for the dilation operation which fills
            in the gaps in the color filter result.

        Returns:
            :obj:`tuple`: a tuple containing:

                :obj:`numpy array`: binary image with the regions of the image that fall in the
                color range

                :obj:`numpy array`: binary image with the edges in the image that fall in the
                color range
        """
        # threshold colors in HSV space
        map = color.inRange(image)
        # binary dilation: fills in gaps and makes the detected regions grow
        kernel = cv2.getStructuringElement(
            cv2.MORPH_ELLIPSE, (dilation_kernel_size, dilation_kernel_size)
        )
        map = cv2.dilate(map, kernel)
        # extract only the edges which come from the region with the selected color
        edge_color = cv2.bitwise_and(map, edges)
        # ---
        return map, edge_color

    @staticmethod
    def find_normal(map: np.ndarray, lines: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        """
        Calculates the centers of the line segments and their normals.

        Args:
            map (:obj:`numpy array`):  binary image with the regions of the image that fall
            in a given color range

            lines (:obj:`numpy array`): An ``Nx4`` array where each row represents a line.
            If no lines were detected,
            returns an empty list.

        Returns:
            :obj:`tuple`: a tuple containing:

                 * :obj:`numpy array`: An ``Nx2`` array where each row represents the center
                 point of a line. If no lines were detected returns an empty list.

                 * :obj:`numpy array`: An ``Nx2`` array where each row represents the normal
                 of a line. If no lines were detected returns an empty list.
        """
        normals = np.array([])
        centers = np.array([])
        if len(lines) > 0:
            length = np.sum((lines[:, 0:2] - lines[:, 2:4]) ** 2, axis=1, keepdims=True) ** 0.5
            dx = 1.0 * (lines[:, 3:4] - lines[:, 1:2]) / length
            dy = 1.0 * (lines[:, 0:1] - lines[:, 2:3]) / length

            centers = np.hstack(
                [(lines[:, 0:1] + lines[:, 2:3]) / 2, (lines[:, 1:2] + lines[:, 3:4]) / 2])
            x3 = (centers[:, 0:1] - 3.0 * dx).astype("int")
            y3 = (centers[:, 1:2] - 3.0 * dy).astype("int")
            x4 = (centers[:, 0:1] + 3.0 * dx).astype("int")
            y4 = (centers[:, 1:2] + 3.0 * dy).astype("int")

            np.clip(x3, 0, map.shape[1] - 1, out=x3)
            np.clip(y3, 0, map.shape[0] - 1, out=y3)
            np.clip(x4, 0, map.shape[1] - 1, out=x4)
            np.clip(y4, 0, map.shape[0] - 1, out=y4)

            flag_signs = (np.logical_and(map[y3, x3] > 0, map[y4, x4] == 0)).astype("int") * 2 - 1
            normals = np.hstack([dx, dy]) * flag_signs

        return centers, normals

    def detect(self, image: BGRImage, colors: List[ColorRange]) -> List[Detections]:
        """
        Detects the line segments in the currently set image that occur in and the edges of
        the regions of the image
        that are within the provided colour ranges.

        Args:
            image (:obj:`numpy array`):         BGR image as numpy array
            colors (:obj:`List[ColorRange]`):   A list of :py:class:`ColorRange` objects specifying
                                                the desired colors to detect.

        Returns:
            :obj:`List[Detections]`:    A list of :py:class:`Detections` objects, one for each
                                        color range given, containing the detected lines for the
                                        corresponding color.

        """
        # find edges
        edges = self.find_edges(image,
                                self.canny_thresholds[0],
                                self.canny_thresholds[1],
                                self.canny_aperture_size)
        # bgr -> hsv
        hsv = cv2.cvtColor(image, cv2.COLOR_BGR2HSV)
        # detect colors
        output: List[Detections] = []
        for color in colors:
            # color filter
            map, edge_color = self.color_filter(hsv, color, edges, self.dilation_kernel_size)
            # hough lines
            lines = self.hough_line(edge_color, self.hough_threshold,
                                    self.hough_min_line_length,
                                    self.hough_max_line_gap)
            # find center and normals
            centers, normals = self.find_normal(map, lines)
            # pack detections
            dets = Detections(lines=lines, normals=normals, map=map, centers=centers)
            output.append(dets)
        # ---
        return output



#!/usr/bin/env python3

from typing import Optional, Dict, List, Tuple

import cv2
import numpy as np

from dt_computer_vision.ground_projection.types import GroundPoint

Color = Tuple[int, int, int]
Segment = Tuple[GroundPoint, GroundPoint]

grid_size = 8
# everything here is calibrated for a 1000x1000 image
scale = 1000
s_segment_thickness = 10
s_padding = 80
s_grid_thickness = 3
s_font_size = 3
s_font_thickness = 3


def draw_grid_image(size: Tuple[int, int]):
    """
    Generates a grid image with distances from the robot's origin.

    Returns:
        :obj:`numpy array`: an OpenCV image

    """
    size_x, size_y = size

    s = max(size_x, size_y) / scale
    padding = int(s_padding * s)
    grid_thickness = max(1, int(s_grid_thickness * s))
    font_size = s_font_size * s
    font_thickness = max(1, int(s_font_thickness * s))

    half_grid_size = int(grid_size / 2)
    cell_size_x = int((size_x - 3 * padding) / grid_size)
    cell_size_y = int((size_y - 3 * padding) / grid_size)
    origin_x, origin_y = 2 * padding + int(grid_size / 2) * cell_size_x, size_y - 2 * padding

    # initialize gray image
    grid_image = np.ones((size_y, size_x, 3), np.uint8) * 128

    # draw vertical lines of the grid
    for col in range(grid_size + 1):
        col_x = col * cell_size_x
        cv2.line(
            grid_image,
            pt1=(2 * padding + col_x, padding),
            pt2=(2 * padding + col_x, origin_y),
            color=(255, 255, 0),
            thickness=grid_thickness
        )

    h_text_centering_offset = int(size_x / 20)
    v_text_centering_offset = int(size_y / 200)

    # draw the horizontal coordinates
    for i in range(0, half_grid_size+1, 2):
        for v in {i, i * -1}:
            text = f"{v * 10}cm"
            text = " " * max(0, 4 - len(text)) + text
            cv2.putText(
                grid_image,
                text,
                (origin_x + v * cell_size_x - h_text_centering_offset,
                 origin_y + h_text_centering_offset),
                cv2.FONT_HERSHEY_PLAIN,
                font_size,
                (255, 255, 255),
                font_thickness,
            )

    # draw horizontal lines of the grid
    for row in range(grid_size + 1):
        row_y = row * cell_size_y
        cv2.line(
            grid_image,
            pt1=(2 * padding, padding + row_y),
            pt2=(size_x - padding, padding + row_y),
            color=(255, 255, 0),
            thickness=grid_thickness
        )

    # draw the vertical coordinates
    for i in range(0, grid_size + 1, 2):
        cv2.putText(
            grid_image,
            f"{i * 10}cm",
            (10, origin_y - i * cell_size_y + v_text_centering_offset),
            cv2.FONT_HERSHEY_PLAIN,
            font_size,
            (255, 255, 255),
            font_thickness
        )

    # draw central vertical line
    col_x = half_grid_size * cell_size_x
    cv2.line(
        grid_image,
        pt1=(2 * padding + col_x, padding),
        pt2=(2 * padding + col_x, origin_y),
        color=(255, 0, 0),
        thickness=grid_thickness
    )

    # draw wheel's axis
    cv2.line(
        grid_image,
        pt1=(2 * padding, origin_y),
        pt2=(2 * padding + grid_size * cell_size_x, origin_y),
        color=(255, 0, 0),
        thickness=grid_thickness
    )

    return grid_image


def debug_image(segments: Dict[Color, List[Segment]], size: Tuple[int, int],
                background_image: Optional[np.ndarray] = None):
    """
    Generates a debug image with all the projected segments plotted with respect to the
    robot's origin.

    Args:
        segments (:obj:`dict`): Line segments in the ground plane relative to robot's origin
        size (:obj:`tuple`): Image of the image to draw
        background_image (:obj:`np.ndarray`): Optional background image

    Returns:
        :obj:`numpy array`: an OpenCV image

    """
    size_x, size_y = size

    s = max(size_x, size_y) / scale
    segment_thickness = max(1, int(s_segment_thickness * s))
    padding = int(s_padding * s)

    half_grid_size = int(grid_size / 2)
    cell_size_x = int((size_x - 3 * padding) / grid_size)
    cell_size_y = int((size_y - 3 * padding) / grid_size)
    origin_x, origin_y = 2 * padding + int(grid_size / 2) * cell_size_x, size_y - 2 * padding

    background_image = background_image if background_image is not None else draw_grid_image(size)
    image = background_image.copy()
    # plot every segment if both ends are in the scope of the image (within 50cm from the origin)
    for color, lines in segments.items():
        for start, end in lines:
            # only draw segments that are within the grid
            if np.any(np.abs([start.x, start.y, end.x, end.y]) > (half_grid_size / 10)):
                continue
            # draw segment
            cv2.line(
                image,
                pt1=(
                    origin_x + int(-(start.y * 10) * cell_size_x),
                    origin_y - int((start.x * 10) * cell_size_y)
                ),
                pt2=(
                    origin_x + int(-(end.y * 10) * cell_size_x),
                    origin_y - int((end.x * 10) * cell_size_y)
                ),
                color=color,
                thickness=segment_thickness,
            )
    # ---
    return image



import dataclasses

import numpy as np
import cv2


@dataclasses.dataclass
class Detections:
    """
    This is a data class that can be used to store the results of the line detection procedure
    performed by :py:class:`LineDetector`.
    """
    # An ``Nx4`` array with every row representing a line ``[x1, y1, x2, y2]``
    lines: np.ndarray
    # An ``Nx2`` array with every row representing the normal of a line ``[nx, ny]``
    normals: np.ndarray
    # An ``Nx2`` array with every row representing the center of a line ``[cx, cy]``
    centers: np.ndarray
    # A binary map of the area from which the line segments were extracted
    map: np.ndarray


class ColorRange:
    """
    The Color Range class holds one or multiple color ranges. It can easily be generated with
    the :py:meth:`fromDict` class method and extends the `OpenCV's inRange <https://docs.opencv.org/3.4/d2/de8/group__core__array.html#ga48af0ab51e36436c5d04340e036ce981>`_
    method to work with multiple color ranges.

    All colours must be given in ``HSV`` space.

    Args:
        low (:obj:`numpy array`): An ``Nx3`` array with the low ends of ``N`` colour ranges.
        high (:obj:`numpy array`): An ``Nx3`` array with the high ends of ``N`` colour ranges.
    """

    def __init__(self, low, high):
        self.low = low
        self.high = high

    @classmethod
    def fromDict(cls, dictionary):
        """

        Generates a :py:class:`ColorRange` object from a dictionary. Expects the colors to be given in ``HSV`` space.
        If multi-entry ranges are provided (e.g. if you are interested in yellow and white), then each should have
        keys ``high_X`` and ``low_X``, see example bellow:

        Examples:

            Single-entry color range::

                { 'low': [0,0,150], 'high': [180,60,255] }

            Multi-entry color range::

                { 'low_1': [0,0,150], 'high_1': [180,60,255], 'low_2': [165,140,100], 'high_2': [180,255,255] }


        Args:
            dictionary (:obj:`dict`): The yaml dictionary describing the color ranges.

        Returns:
            :obj:`ColorRange`: the generated ColorRange object
        """

        # if only two entries: single-entry, if more: multi-entry
        if len(dictionary) == 2:
            assert "low" in dictionary, "Key 'low' must be in dictionary"
            assert "high" in dictionary, "Key 'high' must be in dictionary"
            low = np.array(dictionary["low"]).reshape((1, 3))
            high = np.array(dictionary["high"]).reshape((1, 3))

        elif len(dictionary) % 2 == 0:

            # make the keys tuples with `low` or `high` and the id of the entry
            dictionary = {tuple(k.split("_")): v for k, v in list(dictionary.items())}
            entry_indices = set([k[1] for k, _ in list(dictionary.items())])

            assert len(entry_indices) == len(dictionary) / 2, (
                "The multi-entry definition doesn't " "follow the requirements"
            )

            # build an array for the low and an array for the high range bounds
            low = np.zeros((len(entry_indices), 3))
            high = np.zeros((len(entry_indices), 3))
            for idx, entry in enumerate(entry_indices):
                low[idx] = dictionary[("low", entry)]
                high[idx] = dictionary[("high", entry)]

        else:
            raise ValueError(
                "The input dictionary has two have an even number of "
                "entries: a low and high value for each color range."
            )

        return cls(low=low, high=high)

    def inRange(self, image):
        """
        Applies the `OpenCV inRange
        <https://docs.opencv.org/3.4/d2/de8/group__core__array.html#ga48af0ab51e36436c5d04340e036ce981>`_
        method to every color range entry. Returns the bitwise OR of the results.
        In other words, returns a binary map with 1 for the pixels of the input image that fall
        in at least one of the color ranges.

        Args:
            image (:obj:`numpy array`): an ``HSV`` image

        Returns:
            :obj:`numpy array`: a two-dimensional binary map
        """

        selection = cv2.inRange(image, self.low[0], self.high[0])
        for i in range(1, len(self.low)):
            current = cv2.inRange(image, self.low[i], self.high[i])
            selection = cv2.bitwise_or(current, selection)

        return selection

    @property
    def representative(self):
        """
        Provides an representative color for this color range.
        This is the average color of the first range (if more than one ranges are set).

        Returns:
            :obj:`list`: a list with 3 entries representing an HSV color
        """

        return list(0.5 * (self.high[0] + self.low[0]).astype(int))



import itertools
from typing import Tuple

import numpy as np


def invert_map(mapx: np.ndarray, mapy: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
    H, W = mapx.shape[0:2]
    rmapx = np.empty_like(mapx)
    rmapx.fill(np.nan)
    rmapy = np.empty_like(mapx)
    rmapy.fill(np.nan)

    for y, x in itertools.product(list(range(H)), list(range(W))):
        tx = mapx[y, x]
        ty = mapy[y, x]

        tx = int(np.round(tx))
        ty = int(np.round(ty))

        if (0 <= tx < W) and (0 <= ty < H):
            rmapx[ty, tx] = x
            rmapy[ty, tx] = y

    fill_holes(rmapx, rmapy)
    return rmapx, rmapy


def fill_holes(rmapx, rmapy):
    H, W = rmapx.shape[0:2]

    R = 2
    F = R * 2 + 1

    def norm(x):
        return np.hypot(x[0], x[1])

    deltas0 = [(i - R - 1, j - R - 1) for i, j in itertools.product(list(range(F)), list(range(F)))]
    deltas0 = [x for x in deltas0 if norm(x) <= R]
    deltas0.sort(key=norm)

    holes = set()

    for i, j in itertools.product(list(range(H)), list(range(W))):
        if np.isnan(rmapx[i, j]):
            holes.add((i, j))

    while holes:
        nholes = len(holes)
        nholes_filled = 0

        for i, j in list(holes):
            # there is nan
            nholes += 1
            for di, dj in deltas0:
                u = i + di
                v = j + dj
                if (0 <= u < H) and (0 <= v < W):
                    if not np.isnan(rmapx[u, v]):
                        rmapx[i, j] = rmapx[u, v]
                        rmapy[i, j] = rmapy[u, v]
                        nholes_filled += 1
                        holes.remove((i, j))
                        break

        if nholes_filled == 0:
            break


def ensure_ndarray(obj):
    return obj if isinstance(obj, np.ndarray) else np.array(obj)



import cv2
import numpy as np


def draw_segments(image, detections):
    """

    Draws a set of line segment detections on an image.

    Args:
        image (:obj:`numpy array`): an image
        detections (`dict`): a dictionary that has keys :py:class:`ColorRange` and
        values :py:class:`Detection`

    Returns:
        :obj:`numpy array`: the image with the line segments drawn on top of it.

    """
    im = np.copy(image)
    for color_range, det in list(detections.items()):
        # convert HSV color to BGR
        c = color_range.representative
        c = np.uint8([[[c[0], c[1], c[2]]]])
        color = cv2.cvtColor(c, cv2.COLOR_HSV2BGR).squeeze().astype(int)
        # plot all detected line segments and their normals
        for i in range(len(det.normals)):
            center = det.centers[i]
            normal = det.normals[i]
            im = cv2.line(
                im,
                tuple(center.astype(int)),
                tuple((center + 10 * normal).astype(int)),
                color=(0, 0, 0),
                thickness=2,
            )
            # im = cv2.circle(im, (center[0], center[1]), radius=3, color=color, thickness=-1)
        for line in det.lines:
            im = cv2.line(im, (line[0], line[1]), (line[2], line[3]), color=(0, 0, 0), thickness=5)
            im = cv2.line(
                im, (line[0], line[1]), (line[2], line[3]), color=tuple([int(x) for x in color]),
                thickness=2
            )
    return im


def draw_maps(image, detections):
    """

    Draws a set of color filter maps (the part of the images falling in a given color range)
    on an image.

    Args:
        image (:obj:`numpy array`): an image
        detections (`dict`): a dictionary that has keys :py:class:`ColorRange` and
        values :py:class:`Detection`

    Returns:
        :obj:`numpy array`: the image with the line segments drawn on top of it.

    """
    im = np.copy(image)
    im = cv2.cvtColor(cv2.cvtColor(im, cv2.COLOR_BGR2GRAY), cv2.COLOR_GRAY2BGR)
    color_map = np.zeros_like(im)
    for color_range, det in list(detections.items()):
        # convert HSV color to BGR
        c = color_range.representative
        c = np.uint8([[[c[0], c[1], c[2]]]])
        color = cv2.cvtColor(c, cv2.COLOR_HSV2BGR).squeeze().astype(int)
        color_map[np.where(det.map)] = color
    im = cv2.addWeighted(im, 0.3, color_map, 1 - 0.3, 0.0)
    return im



#!/usr/bin/env python3

import numpy as np

from dt_computer_vision.camera.types import NormalizedImagePoint, CameraModel

from .types import GroundPoint


class GroundProjector:
    """
        Handles the Ground Projection operations.

        This class projects points in the image to the ground plane and in the robot's
        reference frame.
        It enables lane localization in the 2D ground plane.
        This projection uses the homography obtained from the extrinsic calibration procedure.

        Note:
            All pixel and image operations in this class assume that the pixels and images are
            *already rectified*.
            If unrectified pixels or images are supplied, the outputs of these operations will
            be incorrect.

        Args:
            camera (``CameraModel``): Object describing the camera model

        """

    def __init__(self, camera: CameraModel):
        self.camera = camera
        # store homography
        if self.camera.H is None:
            raise ValueError("You need to set a homography `H` in your CameraModel object before "
                             "you can use it to create an instance of GroundProjector.")
        self.H: np.ndarray = self.camera.H.reshape((3, 3))
        # invert homography
        self.Hinv: np.ndarray = np.linalg.inv(self.H)
        # cache/support objects
        self.debug_img_bg = None

    def vector2ground(self, point: NormalizedImagePoint) -> GroundPoint:
        """
        Projects a normalized point (``[-1, 1] X [-1, 1]``) to the ground
        plane using the homography matrix.

        Args:
            point (:py:class:`Point`): A :py:class:`Point` object in normalized coordinates.

        Returns:
            :py:class:`GroundPoint` : A :py:class:`GroundPoint` object on the ground plane.

        """
        point = np.array([point.x, point.y, 1.0])
        ground_point = np.dot(self.H, point)
        ground_point = ground_point / ground_point[2]
        return GroundPoint(*ground_point[:2])

    def ground2vector(self, point: GroundPoint) -> NormalizedImagePoint:
        """
        Projects a point on the ground plane to a normalized pixel (``[-1, 1] X [-1, 1]``) using the
        homography matrix.

        Args:
            point (:py:class:`Point`): A :py:class:`Point` object on the ground plane.

        Returns:
            :py:class:`NormalizedImagePoint` : A point in normalized coordinates.

        """
        ground_point = np.array([point.x, point.y, 1.0])
        image_point = np.dot(self.Hinv, ground_point)
        image_point = image_point / image_point[2]
        return NormalizedImagePoint(*image_point[:2])

    def distorted_to_ground(self, point: NormalizedImagePoint) -> GroundPoint:
        """
        Creates a :py:class:`ground_projection.types.GroundPoint` object from a normalized point
        message from a distorted image. It converts it to pixel coordinates and rectifies it.
        Then projects it to the ground plane.

        Args:
            point (:obj:`ground_projection.types.NormalizedImagePoint`): Normalized point
            coordinates from a distorted image.

        Returns:
            :obj:`ground_projection.types.GroundPoint`: Point coordinates in the ground
            reference frame.

        """
        # rectify point [distorted point -> rectified point]
        point_rect = self.camera.rectifier.rectify_point(point)
        # project on ground [rectified point -> ground point]
        ground_pt = self.vector2ground(point_rect)
        # ---
        return ground_pt

    # def lineseglist_cb(self, seglist_msg):
    #     """
    #     Projects a list of line segments on the ground reference frame point by point by
    #     calling :py:meth:`pixel_msg_to_ground_msg`. Then publishes the projected list of segments.
    #
    #     Args:
    #         seglist_msg (:obj:`duckietown_msgs.msg.SegmentList`): Line segments in pixel space from
    #         unrectified images
    #
    #     """
    #     if self.camera_info_received:
    #         seglist_out = SegmentList()
    #         seglist_out.header = seglist_msg.header
    #         for received_segment in seglist_msg.segments:
    #             new_segment = Segment()
    #             new_segment.points[0] = self.pixel_msg_to_ground_msg(received_segment.pixels_normalized[0])
    #             new_segment.points[1] = self.pixel_msg_to_ground_msg(received_segment.pixels_normalized[1])
    #             new_segment.color = received_segment.color
    #             # TODO: what about normal and points?
    #             seglist_out.segments.append(new_segment)
    #         self.pub_lineseglist.publish(seglist_out)
    #
    #         if not self.first_processing_done:
    #             self.log("First projected segments published.")
    #             self.first_processing_done = True
    #
    #         if self.pub_debug_img.get_num_connections() > 0:
    #             debug_image_msg = self.bridge.cv2_to_compressed_imgmsg(self.debug_image(seglist_out))
    #             debug_image_msg.header = seglist_out.header
    #             self.pub_debug_img.publish(debug_image_msg)
    #     else:
    #         self.log("Waiting for a CameraInfo message", "warn")



from typing import List

import numpy as np

import cv2

from .boards import CalibrationBoard
from ... import CameraModel, Pixel


def estimate_homography(corners: List[Pixel], board: CalibrationBoard,
                        camera: CameraModel) -> np.ndarray:
    """
    Estimates the homography matrix from a list of detected corners and a given known board.

    For more information on the called to perform these operations, consult the OpenCV reference for
    `findChessboardCorners <https://docs.opencv.org/2.4/modules/calib3d/doc
    /camera_calibration_and_3d_reconstruction.html?highlight=findhomography#findchessboardcorners>`_,
    `cornerSubPix <https://docs.opencv.org/2.4/modules/imgproc/doc/feature_detection.html#cornersubpix>`_,
    `findHomography <https://docs.opencv.org/2.4/modules/calib3d/doc
    /camera_calibration_and_3d_reconstruction.html?#findhomographyi>`_.

    Args:
        corners (:obj:``list[Pixel]``): Corners as returned by
            ``dt_computer_vision.camera.calibration.extrinsics.chessboard.find_corners``
        board (:obj:``dt_computer_vision.camera.calibration.extrinsics.boards.CalibrationBoard``):
            The calibration board to use
        camera (:obj:``dt_computer_vision.camera.CameraModel``): Camera model of the camera used

    Returns:
        :obj:``numpy array``: The estimated homography

    Raises:
        RuntimeError: If no corners were found in image, or the corners couldn't be rearranged

    """
    # source corners are on the ground, easily reconstructable given a known board
    src_corners = []
    board_offset = np.array([board.x_offset, board.y_offset])
    square_size = board.square_size
    for r in range(board.rows - 1):
        for c in range(board.columns - 1):
            src_corner = np.array([(r + 1) * square_size, (c + 1) * square_size]) + board_offset
            src_corners.append(src_corner)

    # OpenCV labels corners left-to-right, top-to-bottom, let's do the same
    src_corners = src_corners[::-1]
    # opencv wants corners in an array of shape [N, 1, 2]
    src_corners = np.expand_dims(src_corners, axis=1)

    # NOTE: We're having a problem with our pattern since it's not rotation-invariant

    # destination corners are on the image, in normalized image coordinates
    dst_corners = []
    for corner in corners:
        dst_corner = camera.pixel2vector(corner).as_array()
        dst_corners.append(dst_corner)
    # opencv wants corners in an array of shape [N, 1, 2]
    dst_corners = np.expand_dims(dst_corners, axis=1)

    # compute homography from image to ground using RANSAC
    H, _ = cv2.findHomography(dst_corners, src_corners, cv2.RANSAC)
    # ---
    return H
